/**
 * @file code.h
 * @brief Test file for looking at Undefined tab position LaTeX error
 * @author Mike Kopps
 * @date 12 September 2016
 */

/**
 * @mainpage Interface Layer Overview
 * Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed non molestie
 * nunc, sed interdum mauris. Nunc facilisis, leo ut sodales viverra, augue
 * justo tincidunt lorem, ut efficitur augue magna a leo. Aenean at leo et
 * risus dignissim hendrerit vitae id nibh. Phasellus et laoreet arcu, nec
 * efficitur dolor. Sed tristique lectus ante, quis pulvinar turpis auctor
 * ac. Vestibulum ante ipsum primis in faucibus orci luctus et ultrices
 * posuere cubilia Curae; In consequat sem justo, quis tempus tellus tempor
 * ut. Aenean eget condimentum tellus. Suspendisse et tincidunt arcu. Vivamus
 * a tellus interdum, finibus ipsum ac, tempus risus.
 */

/***************************************************************************
 * NVMe PRP Entry                                                          *
 ***************************************************************************/
/** @brief A physical region page (PRP) entry is a pointer to a physical memory
 *         page. PRPs are used as a scatter/gather mechanism for data transfers
 *         between the controller and memory.  To enable efficient out of order
 *         data transfers between the controller and the host, PRP entries are a
 *         fixed size.
 *
 *  Integer scelerisque egestas metus sed porttitor. Sed eu placerat est. Mauris
 *  ultrices scelerisque lorem mattis luctus. Sed lacinia ex sit amet orci
 *  pretium blandit. Suspendisse potenti. Interdum et malesuada fames ac ante
 *  ipsum primis in faucibus. Etiam magna nulla, porttitor nec egestas ac,
 *  commodo nec sapien. Nam ut turpis ut nisl ultricies iaculis a non mauris.
 *  Aliquam ornare lacinia dolor, at elementum lacus varius eget. Curabitur in
 *  massa id ipsum sodales bibendum. Maecenas vel luctus risus. Etiam tristique
 *  ante magna, ac accumsan lectus facilisis nec. Aliquam eu accumsan dolor,
 *  eget aliquet tellus. Etiam finibus, ipsum in posuere gravida, lectus lacus
 *  tempor odio, vitae semper magna est eget augue. Sed feugiat enim urna, in
 *  viverra lorem elementum et.
 *
 *  Sed quis purus nec felis pellentesque pretium lacinia in felis. Proin quis
 *  porta magna. Sed tortor augue, sollicitudin et orci quis, congue imperdiet
 *  tortor. Nam ultricies magna nec laoreet egestas. Vivamus semper neque eget
 *  elit tristique, ut imperdiet nunc malesuada. Nam in sapien ut mi ultrices
 *  varius sit amet at nulla. Nullam at lacus nec nisl malesuada congue.
 *  Praesent vel lectus accumsan, vestibulum ex sed, varius risus.
 *
 *  Mauris vel volutpat ipsum, ut vulputate lectus. Vivamus accumsan tincidunt
 *  tristique. Cum sociis natoque penatibus et magnis dis parturient montes,
 *  nascetur ridiculus mus. Quisque eu faucibus nibh. Duis posuere consequat
 *  rutrum. Donec quis mollis nisi, in pulvinar magna. Maecenas a tellus id nisl
 *  faucibus vulputate id ac mauris. Praesent suscipit, urna ac lobortis
 *  eleifend, nibh erat varius neque, a convallis libero purus vitae lorem.
 *  Fusce nibh felis, commodo ut metus at, fringilla sollicitudin nunc.
 *
 *  Pellentesque habitant morbi tristique senectus et netus et malesuada fames
 *  ac turpis egestas. Proin volutpat leo vel mauris semper laoreet ut a ligula.
 *  Ut at sem lacus. Donec scelerisque, purus at varius varius, orci ex accumsan
 *  nulla, quis ullamcorper odio lectus et nisl. Nunc massa augue, mattis et
 *  elementum sed, euismod ullamcorper orci. Pellentesque quis arcu eu ante
 *  elementum consequat. Proin in volutpat dolor, ut ornare lacus. Curabitur
 *  rutrum nunc in mauris facilisis, id ultricies nulla posuere. Cras eget
 *  iaculis libero, nec pretium dolor. Sed risus arcu, venenatis nec vehicula
 *  vitae, tempor vitae ligula. Pellentesque et ultrices ex.
 */
typedef struct _PLI_NVME_PRP_ENTRY1
{
    union
    {
        struct
        {
            U32 PBAO_Low                       :  32; /*   0 to  31 */ /**< Page Base Address and Offset - NOTE: Bits 0-1 Reserved  */
            U32 PBAO_High                      :  32; /*  32 to  63 */ /**< Page Base Address and Offset  */
        } bits;
        U64 Qword;                                                     /**< Access via a single U64 variable. */
    } u;                                                               /**< Union of access methods. */
} PLI_NVME_PRP_ENTRY1, *PTR_PLI_NVME_PRP_ENTRY1;

/** @brief A physical region page (PRP) entry is a pointer to a physical memory
 *         page. PRPs are used as a scatter/gather mechanism for data transfers
 *         between the controller and memory.  To enable efficient out of order
 *         data transfers between the controller and the host, PRP entries are a
 *         fixed size.
 */
typedef struct _PLI_NVME_PRP_ENTRY2
{
    union
    {
        struct
        {
            U32 PBAO_Low                       :  32; /*   0 to  31 */ /**< Page Base Address and Offset - NOTE: Bits 0-1 Reserved  */
            U32 PBAO_High                      :  32; /*  32 to  63 */ /**< Page Base Address and Offset  */
        } bits;
        U64 Qword;                                                     /**< Access via a single U64 variable. */
    } u;                                                               /**< Union of access methods. */
} PLI_NVME_PRP_ENTRY2, *PTR_PLI_NVME_PRP_ENTRY2;


/***************************************************************************
 * NVMe SGL                                                                *
 ***************************************************************************/
/**
 * @brief Describes a data block.
 */
typedef struct _PLI_NVME_SGL_DATA_BLOCK_DESC
{
    /** If the SGL Identifier Descriptor Sub Type field is set to 0h then the
     *  Address field specifies the starting 64-bit memory byte address of the
     *  data block. If the SGL Identifier Descriptor Sub Type field is set to 1h
     *  then the Address field contains an offset from the beginning of the
     *  location where data may be transferred. */
    U64    Address;
    /** The Length field specifies the length in bytes of the data block. A Length
     *  field set to 00000000h specifies that no data is transferred. An SGL Data
     *  Block descriptor specifying that no data is transferred is a valid SGL
     *  Data Block descriptor. If the value in the Address field plus the value in
     *  the Length field is greater than 1_00000000_00000000h then the SGL Data
     *  Block descriptor shall be processed as having a Data SGL Length Invalid or
     *  Metadata SGL Length Invalid error.  */
    U32    Length;
    U8     Reserved[3]; /**< Reserved */
    /**
     * Bits[3:0] SGL Descriptor Sub Type field.
     * Bits[7:4] SGL Descriptor Type: 0h @ref PLI_NVME_SGL_DESC_TYPE_DATABLOCK
     */
    U8     Flags;
} PLI_NVME_SGL_DATA_BLOCK_DESC, *PTR_PLI_NVME_SGL_DATA_BLOCK_DESC;

/**
 * @brief The SGL Bit Bucket descriptor is used to ignore parts of source data.
 */
typedef struct _PLI_NVME_SGL_BITBUCKET_DESC
{
    U64    Reserved; /**< Reserved */
    /**
     * The Length field specifies the amount of source data that is discarded. An
     * SGL Bit Bucket descriptor specifying that no source data be discarded (i.e.,
     * the length field set to 00000000h) is a valid SGL Bit Bucket descriptor.
     *
     * If the SGL Bit Bucket Descriptor describes a destination data buffer (e.g., a
     * read from the controller to memory), then the Length field specifies the
     * number of bytes of the source data which the controller shall discard (i.e.,
     * not transfer to the destination data buffer).
     *
     * If the SGL Bit Bucket Descriptor describes a source data buffer (e.g., a
     * write from memory to the controller), then the Bit Bucket Descriptor shall be
     * treated as if the Length field were set to 00000000h (i.e., the Bit Bucket
     * Descriptor has no effect).
     *
     * If SGL Bit Bucket descriptors are supported, their length in a destination
     * data buffer shall be included in the Number of Logical Blocks (NLB) parameter
     * specified in NVM Command Set data transfer commands. Their length in a source
     * data buffer is not included in the NLB parameter.
     */
    U32    Length;
    U8     Reserved1[3]; /**< Reserved */
    /**
     * Bits[3:0] SGL Descriptor Sub Type field.
     * Bits[7:4] SGL Descriptor Type: 1h @ref PLI_NVME_SGL_DESC_TYPE_BITBUCKET
     */
    U8     Flags;
} PLI_NVME_SGL_BITBUCKET_DESC, *PTR_PLI_NVME_SGL_BITBUCKET_DESC;

/**
 * @brief The SGL Segment descriptor describes the next SGL segment, which is
 *        not the last SGL segment.
 */
typedef struct _PLI_NVME_SGL_SEGMENT_DESC
{
    /** If the SGL Identifier Descriptor Sub Type field is set to 0h then the Address
     *  field specifies the starting 64-bit memory byte address of the next SGL
     *  segment, which is a SGL segment. If the SGL Identifier Descriptor Sub Type
     *  field is set to 1h then the Address field contains an offset from the
     *  beginning of the location where data may be transferred.
     */
    U64    Address;
    /** The Length field specifies the length in bytes of the next SGL segment. The
     *  Length field shall be a non-zero value and a multiple of 16.
     *
     *  If the value in the Address field plus the value in the Length field is
     *  greater than 1_00000000_00000000h, then the SGL Segment descriptor shall be
     *  processed as having a Data SGL Length Invalid or Metadata SGL Length Invalid
     *  error.
     */
    U32    Length;
    U8     Reserved[3]; /**< Reserved */
    /**
     * Bits[3:0] SGL Descriptor Sub Type field.
     * Bits[7:4] SGL Descriptor Type: 2h @ref PLI_NVME_SGL_DESC_TYPE_SEGMENT
     */
    U8     Flags;
} PLI_NVME_SGL_SEGMENT_DESC, *PTR_PLI_NVME_SGL_SEGMENT_DESC;

/**
 * @brief The SGL Last Segment descriptor describes the next and last SGL
 *        segment. A last SGL segment that contains an SGL Segment descriptor
 *        or an SGL Last Segment descriptor is processed as an error. */
typedef struct _PLI_NVME_SGL_LAST_SEGMENT_DESC
{
    /** If the SGL Identifier Descriptor Sub Type field is set to 0h then the
     *  Address field specifies the starting 64-bit memory byte address of the next
     *  and last SGL segment, which is a SGL segment. If the SGL Identifier
     *  Descriptor Sub Type field is set to 1h then the Address field contains an
     *  offset from the beginning of the location where data may be transferred. */
    U64    Address;
    /** The Length field specifies the length in bytes of the next and last SGL
     *  segment. The Length field shall be a non-zero value and a multiple of 16.
     *
     *  If the value in the Address field plus the value in the Length field is
     *  greater than 1_00000000_00000000h, then the SGL Last Segment descriptor
     *  shall be processed as having a Data SGL Length Invalid or Metadata SGL
     *  Length Invalid error. */
    U32    Length;
    U8     Reserved[3]; /**< Reserved */
    /**
     * Bits[3:0] SGL Descriptor Sub Type field.
     * Bits[7:4] SGL Descriptor Type: 3h @ref PLI_NVME_SGL_DESC_TYPE_LASTSEGMENT
     */
    U8     Flags;
} PLI_NVME_SGL_LAST_SEGMENT_DESC, *PTR_PLI_NVME_SGL_LAST_SEGMENT_DESC;

#define PLI_NVME_SGL_DESC_TYPE_DATABLOCK   (0x0) /**< NVMe SGL Descriptor Type DataBlock */
#define PLI_NVME_SGL_DESC_TYPE_BITBUCKET   (0x1) /**< NVMe SGL Descriptor Type BitBucket */
#define PLI_NVME_SGL_DESC_TYPE_SEGMENT     (0x2) /**< NVMe SGL Descriptor Type Segment */
#define PLI_NVME_SGL_DESC_TYPE_LASTSEGMENT (0x3) /**< NVMe SGL Descriptor Type Last Segment */

#define PLI_NVME_SGL_DESC_FLAGS_DESCTYPE_SHIFT (0x4) /**< NVMe SGL Descriptor Flag DescriptorType field shift */
#define PLI_NVME_SGL_DESC_FLAGS_DESCTYPE_MASK  (0xF0) /**< NVMe SGL Descriptor Flag DescriptorType field mask */

/** @brief A union of all of the available NVMe SGL Descriptor formats. */
typedef union _PLI_NVME_SGL_DESCRIPTOR
{
    PLI_NVME_SGL_DATA_BLOCK_DESC      DataBlockDesc;   /**< NVMe SGL Descriptor Type DataBlock */
    PLI_NVME_SGL_BITBUCKET_DESC       BitBucketDesc;   /**< NVMe SGL Descriptor Type BitBucket */
    PLI_NVME_SGL_SEGMENT_DESC         SegmentDesc;     /**< NVMe SGL Descriptor Type Segment */
    PLI_NVME_SGL_LAST_SEGMENT_DESC    LastSegmentDesc; /**< NVMe SGL Descriptor Type Last Segment */

} PLI_NVME_SGL_DESCRIPTOR, *PTR_PLI_NVME_SGL_DESCRIPTOR;
